-- GSI - general sound interface
-- sound.e
--
-- Copyright (C) W.H.Scholten 1996,1997,1998
--
-- ** See gsi.h for a description of the commands **
--
-- License: CGPL (see the file LICENSE.CGPL)
--

class GSI_INTERFACE
   --Note that this was based on SND_SERVER 0.47 and is being revised 
   --and updated for GSI

inherit
   LIBC

   GSI_CONFIG -- equivalent of config.h for C
--char *mixerdev=MIXERDEV;
--static int  soundfd;
--static char sndserver_ok = 1;
--static char mixer_ok = 1;
--
--int  sound_volume=4;
--int  sound_on=1;

   num_commands: INTEGER
   sound_command: ARRAY[CHARACTER]

   MAX_QUEUE_LENGTH : INTEGER is 256

feature
   GSI_master_vol, GSI_pcm_vol, GSI_synth_vol, GSI_cd_vol: INTEGER

   server_id: INTEGER -- process id of the gsi_server

   channel_status: BIT 256

   cb_function: ARRAY[GSI_CALLBACK_FUNCTION]
   error_callback_function: GSI_ERROR_FUNCTION

   child_killed(i: INTEGER) is
      do
	 debugmsg("child killed, no ", i)
	 sndserver_ok:=0
	 io.put_string("sound server killed%N")
      end

   --*-----------------------------------------------------------------------*
   --	GSI initialisation
   --	here read the hardware master volume, pcm volume, synth volume cd volume
   --	Or do this with a different piece of code above? (init volumes at -1)
   --	Forking the server should also go in here.
   --*-----------------------------------------------------------------------*
   gsi_init: INTEGER is
      local mixer: INTEGER
      do
	 !!send_command(0, MAX_QUEUE_LENGTH)
	mixer := open(mixerdev, O_WRONLY + O_NONBLOCK);
	if (mixer==-1) then
	   mixer_ok:=0
	   err.put_string("GSI: Couldn't open mixer ")
	   err.put_string(mixerdev)
	   err.put_newline
	else
	   ioctl(mixer, SOUND_MIXER_READ_DEVMASK, $mixer_devices)
	   close(mixer)
	end
	if ((mixer_devices.to_bit & SOUND_MASK_VOLUME)=0B)
	   emulate_master_volume:=True
	end
	Result:=0
     end


     --*---------------------------------------------------------------------*
     --	Set sound volumes
     --  Mixer volume settings:
     --    VOLUME : master volume, i.e. headphone and line out.
     --    PCM    : dsp volume
     --    SYNTH  : onboard midi playing device
     --
     --  saving volumes is done to remove any rounding problems with repeated
     --  reading/writing (device values read may not be the same as what 
     --  has been set)
     --*---------------------------------------------------------------------*
   gsi_set_master_volume(vol: INTEGER) is
      require vol>=0 and vol <257
      do
	 GSI_master_vol:=vol
	 if emulate_master_volume then
	    gsi_set_pcm_volume(GSI_pcm_vol)
	    gsi_set_synth_volume(GSI_synth_vol)
	    gsi_set_cd_volume(GSI_cd_vol)
	 else
	    set_volume(SOUND_MIXER_WRITE_VOLUME, vol)
	 end
      end

   gsi_set_pcm_volume(vol: INTEGER) is
      require vol>=0 and vol <257
      do
	 GSI_pcm_vol:=vol
	 if emulate_master_volume then
	    vol:=(vol*GSI_master_vol)//256
	    set_volume(SOUND_MIXER_WRITE_PCM, vol)
	 end
      end
   gsi_set_synth_volume(vol: INTEGER) is
      require vol>=0 and vol <257
      do
	 GSI_synth_vol:=vol
	 if emulate_master_volume then
	    vol:=(vol*GSI_master_vol)//256
	    if (mixer_devices.to_bit & SOUND_MASK_SYNTH)=0B then
	       set_volume(SOUND_MIXER_WRITE_SYNTH, vol);
	    end
	 end
      end

   gsi_set_cd_volume(vol: INTEGER) is
      require vol>=0 and vol <257
      do
	 GSI_cd_vol:=vol
	 if emulate_master_volume then
	    vol:=(vol*GSI_master_vol)//256
	    set_volume(SOUND_MIXER_WRITE_CD, vol)
	 end
      end

   gsi_get_master_volume : INTEGER is
      do
	 if GSI_master_vol<0 then
	    GSI_master_vol:=get_volume(SOUND_MIXER_READ_VOLUME)
	 end
	 Result:=GSI_master_vol
      end

   gsi_get_pcm_volume : INTEGER is
      do
	if GSI_pcm_vol<0 then
	   GSI_pcm_vol:=get_volume(SOUND_MIXER_READ_PCM)
	end
	Result:=GSI_pcm_vol
     end

     gsi_get_synth_volume: INTEGER is
      do
	if GSI_synth_vol<0 then
	   if (mixer_devices & SOUND_MASK_SYNTH)
	      GSI_synth_vol=get_volume(SOUND_MIXER_READ_SYNTH);
	   end
	end
	Result:=GSI_synth_vol
     end

     gsi_get_cd_volume: INTEGER is
      do
	 if (GSI_cd_vol<0) then
	    GSI_cd_vol=get_volume(SOUND_MIXER_READ_CD);
	 end
	 Result:=GSI_cd_vol
      end


   -- We expect volumes in the range 0-255
   -- (Mono at the moment, channel L and R are both set to the 
   -- requested value)
   set_volume(device, vol: INTEGER) is
      local
	 mixer: INTEGER
      require vol>=0 and vol <256
      do
	 if mixer_ok then
	    mixer:= open(mixerdev, O_WRONLY + O_NONBLOCK)
	    volume:=(vol*100)//256 --OSS uses a scale 0-100
	    volume:= volume + volume @<< 8 -- LEFT AND RIGHT CHANNEL >>
	    ioctl(mixer, device, $volume)
	    close(mixer)
	 end
      end

   -- We return volumes in the range 0-255
   -- (Mono at the moment, only channel L is returned)
   get_volume(device): INTEGER) is
      local
	 mixer: INTEGER
	 volume_L: INTEGER
      do
	 if mixer_ok then
	    mixer:= open(mixerdev, O_RDONLY + O_NONBLOCK)
	    ioctl(mixer, device, $volume)
	    volume_L := volume &255
--	    volume_R := volume//256
--	    volume :=(volume_L*255/100) | ((volume_L*255/100)<<8) //>>>>
	    volume :=(volume_L*255)/100
	    close(mixer)
	 end
      end

   --*-----------------------------------------------------------------------*
   --*	Init PCM system (temporarily also forks the server)
   --*-----------------------------------------------------------------------*
   gsi_init_pcm(dsp_speed, dsp_channels, dsp_bytes, dsp_fragsize: INTEGER; sounddir: STRING; soundfile: STRING) is
      local status: INTEGER
      do
	 status:=fork_server

	 if sndserver_ok then
--	//server starts PAUSE'd, DSP released.
	    send_byte(SND_SET_DSP_SPEED);   send_int16(dsp_speed);
	    send_byte(SND_SET_CHANNELS);    send_byte(dsp_channels);
	    send_byte(SND_SET_BYTES);       send_byte(dsp_bytes);
	    send_byte(SND_SET_FRAGSIZE);    send_byte(dsp_fragsize);
	    send_byte(SND_SET_SOUND_DIR);   send_string(sounddir);
	    send_byte(SND_LOAD_SOUND_LIST); send_string(soundfile);
	    send_byte(SND_LOAD_LISTED);
	    send_byte(SND_GRAB_DSP); --This sets the new parameters.
	 end

	 io.put_string("Sound server initialized.%N")
      end



   fork_server : INTEGER is
      local
	 sndserver_filename: POINTER
      do
--    int  child, fd[2];

	 !!sound_commands.make(0,255)

	 sndserver_filename:=getenv("DSPSERVER")

--    if (!(sndserver_filename=getenv("DSPSERVER"))) sndserver_filename=BINDIR"/"SOUNDSERVER;
	 if sndserver_filename=NULL then sndserver_filename:=(BINDIR+"/"+SOUNDSERVER).to_external end
--//    strcpy(sndserver_filename, BINDIR"/"SOUNDSERVER);
	 
	 signal(SIGCHLD, $child_killed)

	 -- Create a pipe, set the write end to close when we exec the 
	 -- sound server, and set both ends to non-blocking (in the server)
	 pipe(fd)
	 soundfd:=fd @ 1
	 --Feedback:
	 pipe(fd2);
	 soundfd2=fd2 @ 0

	 child:=fork

	 if child=0 then --child's thread
	    close(fd @ 1)
	    dup2(fd[0], STDIN_FILENO)
	    --	dup2(fd[0], STDIN_FILENO);
	    --	close(fd[0]);
	    close(fd @ 0)

	    --	// close(fd2[0]);
	    close(fd2 @ 0)
	    --	// dup2(fd2[1], STDOUT_FILENO);
	    dup2(fd @ 1, STDOUT_FILENO)
	    --	// close(fd2[1]);
	    close(fd2 @ 1)

	    --create a C argument list:
	    --	argv[0] = sndserver_filename;
	    --	argv[1] = "-loop";
	    --	argv[2] = NULL;
	    --	execvp(sndserver_filename, argv);
	    execvp(sndserver_filename.to_external, argv)
	    -- We only arrive here if execvp failed to replace this process
	    -- with the soundserver code

	    stderr.put_string("Couldn't execute gsi_server%N")
	    die_with_code(0)
	 end

--    close(fd[0]);
--    // close(fd2[1]);
	 close(fd @ 0)
	 close(fd2 @ 1)

--    DEB1(printf("sndserver_ok=%d\n", sndserver_ok));
	 Result:=sndserver_ok
      end

   --*-----------------------------------------------------------------------*
   --	Init SYNTH system
   --*-----------------------------------------------------------------------*
   gsi_init_synth (preferred_type: INTEGER): INTEGER is
      do
	stderr.put_string("GSI: gsi_init_synth is not yet implemented%N")
	Result:= -1
      end


   --*-----------------------------------------------------------------------*
   --*	Play sounds 'n stuff
   --*-----------------------------------------------------------------------*
   gsi_kill_sounds is
      do
	 num_commands:=0
	 send_byte(SND_KILLALL)
      end


   -- Send a list of commands with data (bytes):
   gsi_send_commands(list: ARRAY[INTEGER]) is
      local
	 i: INTEGER
      do
	 from i:=1 until i>list.size loop
	    sound_command.put(list.item(i), num_commands)
	    num_commands :=num_commands+1
	    if num_commands=MAX_QUEUE_LENGTH then gsi_flush end
	    i:=i+1
	 end
      end


   gsi_play_sound (slot, command: INTEGER) is
      do
	 if sound_on and sndserver_ok then
	    if num_commands==MAX_QUEUE_LENGTH-1 then gsi_flush end
	    sound_command.put(command, num_commands)
	    num_commands:=num_commands+1
	    sound_command.put(slot, num_commands)
	    num_commands:=num_commands+1
	    debug
	       io.put_string("play sound (snd ") io.put_integer(k)
	       io.put_string(", com ") io.put_integer(command)
	       io.put_string(", num commands now ") io.put_integer(num_commands)
	       io.put_new_line
	    end
	 end
      end


   gsi_play_channel_sound (channel, slot, command: INTEGER) is
      require channel>=0 and channel <256
      do
	 if sound_on and sndserver_ok then
	    if num_commands==MAX_QUEUE_LENGTH-2 then gsi_flush end
	    sound_command.put(command + SND_CHANNEL, num_commands)
	    num_commands:=num_commands+1
	    sound_command.put(channel, num_commands)
	    num_commands:=num_commands+1
	    sound_command.put(slot, num_commands)
	    num_commands:=num_commands+1
	 end
      end


   gsi_play_sound_3d (ch, slot, command, x, y, z, dx, dy, dz: INTEGER; doppler_on: BOOLEAN) is
      require ch>=-1 and ch<256
      local channel_flag: INTEGER
      do
	 channel_flag:=0
	 if ch /= -1 then channel_flag=SND_CHANNEL end
	 if sound_on then
	    gsi_send_commands(<<(command.to_bit or channel_flag.to_bit).to_integer>>)
		if ch /= -1 then gsi_send_commands(<<ch>>) end
		gsi_send_commands(<<slot,
				    SND_SET_X + SND_CURRENT + channel_flag, x \\256, x //256,
				    SND_SET_Y + SND_CURRENT + channel_flag, y \\256, y //256,
				    SND_SET_Z + SND_CURRENT + channel_flag, z \\256, z //256,
				    SND_SET_DX + SND_CURRENT + channel_flag, dx \\256, dx //256,
				    SND_SET_DY + SND_CURRENT + channel_flag, dy \\256, dy //256,
				    SND_SET_DZ + SND_CURRENT + channel_flag, dz \\256, dz //256
				    >>)
		if (doppler_on) gsi_send_commands(1, SND_DOPPLER_ON | channel_flag | SND_CURRENT);
	}
	 end
      end

   gsi_play_sound_stereo (channel, slot, command, x, dx: INTEGER) is
      do
	 if sound_on and sndserver_ok then
	    if channel<0 then
	       general_sound(<<command, slot, SND_SET_X + SND_CURRENT, (x.to_bit and 255).to_integer, (x.to_bit @>>8).to_integer, SND_SET_DX + SND_CURRENT, dx>>)
	    else
	       general_sound(<<command, channel, slot, SND_SET_X + SND_CURRENT + SND_CHANNEL, (x.to_bit and 255).to_integer, (x.to_bit @>>8).to_integer, SND_SET_DX + SND_CURRENT +SND_CHANNEL, dx>>)
	    end
	 end
      end

   gsi_play_sound_pan (ch, slot, command, pan: INTEGER) is
      require
	 ch>=-1 and ch <256
	 pan <=256 and pan >=-256
      local channel_flag: INTEGER
      do
	 if ch=-1 then channel_flag:=0 else channel_flag:=SND_CHANNEL end
	 if sound_on then
	    gsi_send_commands(<<(command.to_bit or channel_flag.to_bit).to_integer>>)
	    if ch /= -1 then gsi_send_commands(<< ch>>) end
	    gsi_send_commands(<<slot,
				SND_SET_PAN  + SND_CURRENT + channel_flag, pan \\256, pan //256
				)
	 end
      end


   --*-----------------------------------------------------------------------*
   --*	Play music
   --*-----------------------------------------------------------------------*
   gsi_play_song(loopno: INTEGER) : INTEGER is
      require loopno>=0 and loopno <256
      do
	gsi_send_commands(2, SND_PLAY_SONG, loopno)
	Result:=0
      end
   gsi_stop_song: INTEGER is
      do
	gsi_send_commands(1, SND_STOP_SONG)
	Result:=0
      end
   gsi_continue_song: INTEGER is
      do
	gsi_send_commands(1, SND_CONTINUE_SONG)
	Result:=0
      end




   gsi_flush is
      local
	 i: INTEGER
      do
	 from i:=0 until i=num_commands loop
	    send_byte(sound_command @ i)
	 end
	 num_commands:=0
      end

   gsi_play_sound_flush (slot, command: INTEGER) is
      do
	 gsi_play_sound (slot, command)
	 gsi_flush()
      end

   gsi_pause_sound is
      do
	 if sndserver_ok then send_byte(SND_PAUSE_ALL) end
      end

   gsi_continue_sound is
      do
	 if sndserver_ok then send_byte(SND_CONTINUE_ALL) end
      end

   gsi_grab_pcm is
      do
	 if sndserver_ok then send_byte(SND_GRAB_DSP) end
      end

   gsi_release_pcm is
      do
	 if sndserver_ok then send_byte(SND_RELEASE_DSP) end
      end

   gsi_close_pcm is
      do
	 release_sound
	 if sndserver_ok then
	    send_byte(SND_CLOSE_SERVER)
	 end
	 sndserver_ok:=False
      end

   gsi_close_synth is
      do
      end

   --*-----------------------------------------------------------------------*
   --*	Callbacks
   --*-----------------------------------------------------------------------*
   status_queue : ARRAY[CHARACTER]
   status_queue_len: INTEGER

   b: CHARACTER

   _gsi_check_callback is
      local
	int  i
	int  decoded_length=0
      do
	debugmsg("GSI: callback check\n")
	if sndserver_ok then
	from until not data_available loop
				i:=read(statusfd, $b, 1)
		-- DEB1(fprt(stderr, " -> read status=%d\n", i));
		if i=0 then data_available:=False -- EOF on pipe, i.e. child died. Just quit and let the handler disable sound.
		-- DEB1(if (i<0) fprintf(stderr, "GSI: callback errno=%d\n", errno));
		elseif i<0 then
		   if (errno==EPIPE) then sndserver_ok=0 end
		elseif i>0 then --there was something
				-- DEB1(fprintf(stderr, "GSI: callback\n"))
				status_queue.put(b, status_queue_len)
				status_queue_len:=status_queue_len+1
--			int command=b; //=> if (command_size==0) command=b;
		   inspect status_queue @ 0
		   when  GSI_CALLBACK_OK then
				decoded_length:=1
		   when GSI_CALLBACK_ERROR then -- general error
				decoded_length:=1
		   when GSI_CALLBACK_NO_SUCH_FILE then
				decoded_length:=1
				-- Should also get the slot number
--				fprintf(stderr, "no such file\n");
		      if error_callback_function /=Void then
			  error_callback_function(GSI_CALLBACK_NO_SUCH_FILE)
		      end
		   when GSI_CALLBACK_BAD_LIST then -- eg sounds that can't be loaded ??
		   when GSI_CALLBACK_BAD_COMMAND then
		when GSI_CALLBACK_DSP_BUSY then
			when GSI_CALLBACK_SYNTH_BUSY then
				decoded_length:=1
--			when GSI_CALLBACK_END then -- useless for the stack, unless we add sample id's generated when starting a sound, but must use callbacks for that too. Not worth it I think
			when GSI_CALLBACK_CHANNEL_END then
				int  j;
				if (status_queue_len<2) continue;
				decoded_length=2;
				j=status_queue[1];
				channel_status[j>>5] &=~(1<<(j &31));
				DEB1(fprt(stderr, "GSI_CALLBACK_CHANNEL_END: ch=%d\n", j));
				if (cb_function[j+1]!=NULL) (cb_function[j+1])(j);
			}
			break;

--  why is this needed? Suppose a sound is killed or ends, a play_channel_sound is given and gsi_set_channel_status(ch, 1) given, then a kill CALLBACK can come after this and channel status will be 0 even though the sound has just started. This means we need either a counter (increase with gsi_adjust_channel_status(ch) which increses it by 1, CALLBACK decreases it by 1, or callbacks (which will arrive in the same order as things change in the server)
--KILLALL must give a callback to reset all counters to 0
 
			when GSI_CALLBACK_CHANNEL_STARTED then
			{
				int  j;
				if (status_queue_len<2) continue;
				decoded_length=2;
				j=status_queue[1];
				channel_status[j>>5] |=(1<<(j&31));
				DEB1(fprt(stderr, "GSI_CALLBACK_CHANNEL_STARTED: ch=%d\n", j));
			when GSI_CALLBACK_CHANNEL_ENDALL then
				decoded_length=1;
				int  j;
				for (j=0; j<256; j++) {
					if (_gsi_get_channel_status(j) && cb_function[j+1]!=NULL)
						(cb_function[j+1])(j);
					_gsi_set_channel_status(j, 0);
				}
//				for (j=0; j<8; j++) channel_status[j] = 0;
				DEB1(fprt(stderr, "GSI_CALLBACK_CHANNEL_ENDALL\n"));
--			when GSI_CALLBACK_SONG_START then
--needed ?
			when GSI_CALLBACK_SONG_END then
				decoded_length=1;
				DEB1(fprt(stderr, "GSI_CALLBACK_SONG_END\n"));
				if (cb_function[0]!=NULL) (cb_function[0])(-1);

			else 
			   fprintf(stderr, "GSI: Unkown callback %d\n", status_queue[0]);
			decoded_length=1; /* remove the byte */
			end
		if decoded_length>0 then
		   status_queue_len = status_queue_len- decoded_length
		   if (status_queue_len>0) then
			   memmove(status_queue, status_queue+decoded_length, status_queue_len);
		   end
		end
	     end
      end

   gsi_set_channel_callback (ch: INTEGER; cbf: GSI_CALLBACK_FUNTCION) : GSI_CALLBACK_FUNTCION is
      require ch>=-1 and ch<256
      do
	Result:= cb_function @ (ch+1)

	cb_function.put(cbf, ch+1)
      end
   gsi_clear_channel_callback (ch: INTEGER) is
      require ch>=-1 and ch<256
      do
	cb_function.put(Void, ch+1)
      end

   gsi_set_error_callback (cbf: GSI_ERROR_FUNTCION) : GSI_ERROR_FUNTCION is
      do
	Result:= error_callback_function

	error_callback_function:=cbf
      end
   gsi_clear_error_callback is
      do
	error_callback_function:=Void
      end

   --Return the file descriptor for callbacks so we can use select on it
   gsi_callback_fd : INTEGER is
      do
	 Result:=statusfd
      end

   --For eiffel, use a bit array.
   _gsi_get_channel_status (ch: INTEGER) : INTEGER is
      require ch>=-1 and ch <256
      do
--	 Result:= channel_status[(ch) @>>5] & (1@<< ((ch)&31));
	Result:=(channel_status @ ch) =1B
      end

   _gsi_set_channel_status (ch, bit: INTEGER) is
      require ch>=-1 and ch <256
      do
--	channel_status[ch >>5] = (channel_status[ch >>5] & ~(1<< (ch &31))) | ( (bit &1) << (ch &31) 
--	);
        channel_status.put(0B, ch)
      end

   gsi_get_channel_status (ch: INTEGER) : INTEGER is
      do
        _gsi_check_callback(); --handle any data on the input pipe
--	return channel_status[ch >>5] & (1<< (ch&31));
	Result:=(channel_status @ ch) =1B
      end

   -- + -------------------------------------------------------- +
   --Pipe functions
   pipe_byte: CHARACTER

   send_byte(data: BIT 32) is
      do
	 pipe_byte:=data and 255
	 debugmsg("sound byte: sending ", pipe_byte)
	 if sndserver_ok then write (soundfd, &pipe_byte, sizeof (byte)) end
      end

   send_int16(data: BIT 32) is
      do
	 if sndserver_ok then
	    send_byte(data)
	    send_byte(data @>>8)
	 end
      end

   send_int32(data: BIT 32) is
      local
      do
	 if sndserver_ok then
	    send_byte(data)
	    send_byte(data @>>8)
	    send_byte(data @>>16)
	    send_byte(data @>>24)
	 end
      end

   send_string(s: STRING) is
--      local
--	 n: INTEGER
      do
	 if sndserver_ok then
	    if s=Void or else s.size=0 then
	       send_byte(0)
	    else
--	    n:=write (soundfd, s.to_external, s.length)
	       send_block(s.to_external, s.size)
--	       debugmsg("sound string: no bytes = ", n)
	 end
      end

   gsi_send_block(data: POINTER; len: INTEGER) is
      require len>0
      local
	 write_fd_sel: FD_SET
	 i, j: INTEGER
      do
	 i:=0
	 j:=0
--	DEB1(prt("GSI: _gsi_send_block: sending a block of %d bytes\n", len))
	if sndserver_ok then
	   from until i=len loop
	      j:=write (soundfd, data+i, len-i))
	      if (j>0) then i =i+j end
--			FD_ZERO($write_fd_sel)
--			fd_FD_SET(soundfd, $write_fd_sel)
	      write_fd_sel.FD_ZERO
	      write_fd_sel.FD_SET(soundfd)
--			select(soundfd+1, NULL, $write_fd_sel, NULL, NULL)
	      fd_select(soundfd+1, NULL, write_fd_sel.data, NULL, NULL)
	   end
	end
     end
end
