/* GSI - general sound interface   [-*- C++ -*-]
 *
 * clients.c - shared sound server, client management
 *   - sept 1996/may 1997/dec 1997/march-may 1998
 *
 * Copyright (C) 1996,1997,1998 W.H.Scholten
 *
 * Lower level API dependencies:
 *   - DSP mode setting: OSS but should not be too difficult to adapt
 *     to other systems.
 *   - OSS include locations are Linux/FreeBSD specific.
 *     Locations for other OS's?
 *   - sequencer/synth (MIDI's) playing: very OSS specific.
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */




/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <errno.h>

#include <gsi/config.h>
#define USE_SOCKETS

#include <gsi/mytypes.h>
#include <gsi/debug.h>
#define _SNDSERVER_
#include <gsi/gsi_commands.h>
#include <gsi/gsi_server_version.h>
#include "gsi_server.h"
#include "midi.h"

#ifdef USE_SOCKETS
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <time.h>
#endif /* USE_SOCKETS */


/*__________________DEFINITIES___________________*/

#define min(x,y) ((x)>(y) ? (y): (x))
#define max(x,y) ((x)<(y) ? (y): (x))


/*__________________EXTERN___________________*/
extern char *server_name;
extern int  dsp;

/* gsi_server.c: */
extern struct client_data *current;

/* load.c: */
extern void  unload_sound(int slot);

/*__________________VARIABELEN___________________*/
int  num_clients;
struct client_data  *client_list[64];
/* 64 clients max , or do struct client_data *next_client; struct client_data *prev_client; in the client_data struct to form a two way linked list */

#ifdef USE_SOCKETS
int  ls;				/* listen socket descriptor */

struct hostent *hp;		/* pointer to host info for remote host */
struct servent *sp;		/* pointer to service information */

long   timevar;			/* contains time returned by time() */


struct linger linger;	/* allow a lingering, graceful close; */
						/* used when setting SO_LINGER */

struct sockaddr_in myaddr_in;	/* for local socket address */
struct sockaddr_in peeraddr_in;	/* for peer socket address */
#endif /* USE_SOCKETS */

/*__________________FUNCTIE PROTOTYPES___________________*/

/*__________________PROGRAMMA___________________*/


/*---------------------------------------------------------------------------*
 *	set up the listening socket etc.
 *---------------------------------------------------------------------------*/
int init_connection() {
	DEB1(fprt(stderr, "init_connection\n"));
#ifdef USE_SOCKETS
	/* ==================================================================== */
	/* clear out address structures */
	memset ((char *)&myaddr_in, 0, sizeof(struct sockaddr_in));
	memset ((char *)&peeraddr_in, 0, sizeof(struct sockaddr_in));

	/* Set up address structure for the listen socket. */
	myaddr_in.sin_family = AF_INET;
	/* The server should listen on the wildcard address,
	 * rather than its own internet address.  This is
	 * generally good practice for servers, because on
	 * systems which are connected to more than one
	 * network at once will be able to have one server
	 * listening on all networks at once.  Even when the
	 * host is connected to only one network, this is good
	 * practice, because it makes the server program more
	 * portable.
	 */
	myaddr_in.sin_addr.s_addr = INADDR_ANY;
	sp = getservbyname ("gsi_server", "tcp");
	if (sp == NULL) {
		fprintf(stderr, "%s: gsi_server not found in /etc/services\n", server_name);
		exit(1);
	}
	myaddr_in.sin_port = sp->s_port;

	/* Create the listen socket. */
	ls = socket (AF_INET, SOCK_STREAM, 0);
	if (ls == -1) {
		perror(server_name);
		fprintf(stderr, "%s: unable to create socket\n", server_name);
		exit(1);
	}
	/* Bind the listen address to the socket. */
	if (bind(ls, (struct  sockaddr *) &myaddr_in, sizeof(struct sockaddr_in)) == -1) {
		perror(server_name);
		fprintf(stderr, "%s: unable to bind address\n", server_name);
		exit(1);
	}

	/* Initiate the listen on the socket so remote users can connect. 
	 * The listen backlog is set to 5, which is the largest currently
	 * supported (BSD at least). */
	if (listen(ls, 5) == -1) {
		perror(server_name);
		fprintf(stderr, "%s: unable to listen on socket\n", server_name);
		exit(1);
	}
	DEB1(fprt(stderr, "listening , ls=%d\n", ls));

	/* Set the 'close on exec' flag, not really needed at the moment */
    if ( fcntl(ls, F_SETFD, FD_CLOEXEC ) < 0 ) {
		fprintf(stderr,"Unable to set socket to 'close on exec'\n");
		return( -1 );
	}
    if (fcntl(ls, F_SETFL, O_NONBLOCK) < 0 ) {
		fprintf(stderr,"Unable to set socket to non-blocking\n");
		return( -1 );
    }
	/* ==================================================================== */
#endif /* USE_SOCKETS */
	return 0;
}


int add_client() {
	int  i;
	int  addrlen;
	char *hostname;		/* points to the remote host's name string */
	struct client_data  *new;
	int  socket_fd;		/* connected socket descriptor */

	if (num_clients>=64) return -1;

#ifdef USE_SOCKETS
	/* ==================================================================== */
	addrlen = sizeof(struct sockaddr_in);

	/* This call will block until a new connection arrives unless ls was set
	 * to non-blocking.
	 * Then, it will return the address of the connecting peer, and a new
	 * socket descriptor, socket_fd, for that connection. */
	socket_fd = accept(ls, (struct sockaddr *) &peeraddr_in, &addrlen);

	DEB1(fprt(stderr, "add_client: socket_fd=%d\n", socket_fd));

	if ( socket_fd == -1) return -1;

	/* enable local address reuse */
    {int  curstate = 1;
    if ( setsockopt( socket_fd, SOL_SOCKET, SO_REUSEADDR,
					 &curstate, sizeof(curstate) ) < 0 ) {
		fprintf(stderr,"Unable to set for a fresh socket\n");
		return( -1 );
	}
	}

	/* Look up the host information for the remote host that we have connected
	 * with.  Its internet address was returned by the accept call, in the main
	 * daemon loop above. */
	hp = gethostbyaddr ((char *) &peeraddr_in.sin_addr,
				sizeof (struct in_addr),
				peeraddr_in.sin_family);

	if (hp == NULL) {
		/* The information is unavailable for the remote host.
		 * Just format its internet address to be printed out in the logging
		 * information.  The address will be shown in "internet dot format". */
		hostname = inet_ntoa(peeraddr_in.sin_addr);
	} else {
		hostname = hp->h_name;	/* point to host's name */
	}

	/* Log a startup message. */
	time (&timevar);
	fprintf(stderr, "gsi_server: startup from %s port %u at %s",
		hostname, ntohs(peeraddr_in.sin_port), ctime(&timevar));

	/* Set the socket for a lingering, graceful close.
	 * This will cause a final close of this socket to wait until all of the
	 * data sent on it has been received by the remote host. */
#if 1
	linger.l_onoff  =1;
	linger.l_linger =1;
	if (setsockopt(socket_fd, SOL_SOCKET, SO_LINGER, &linger,
					sizeof(linger)) == -1) {
		fprintf(stderr, "Connection with %s aborted on error\n", hostname);
		exit(1);
	}
#endif
	/* ==================================================================== */
#endif /* USE_SOCKETS */



	/* Now we have a connection, so add a new client */
	client_list[num_clients] = (struct client_data *) calloc(1, sizeof(struct client_data));
	if (client_list[num_clients] == NULL) return -1; /* malloc failed */

	new=client_list[num_clients];
	num_clients++;

	/* initialize data */
	new->mach1=256<<8;
	new->max_stack=MAX_DYNAMIC_CHANNELS;
	new->base_width=256;
	new->last_sound_stack = -1;
	new->last_sound_channel =-1;

	new->new_dsp_speed=8000;
	new->new_dsp_log2_frag=8;
	new->new_dsp_channels=1;
	new->new_dsp_type=GSI_8BIT;
	new->dsp_request=SND_RELEASE_DSP; /* indicating the DSP is not (yet) requested */
	new->socket_fd=socket_fd;

	for (i=0; i<MAX_STATIC_CHANNELS; i++)
		new->playing[i+MAX_DYNAMIC_CHANNELS]=-1;

	fcntl(/*new->*/socket_fd, F_SETFL, O_NONBLOCK);

	return 0;
}


int remove_client(int client) {
	int  i;

	if ((client &255)<num_clients) {
		close(client_list[client]->socket_fd);

		/* free client data */
		current=client_list[client];
		for (i=0; i<256; i++) {
			unload_sound(i);
			free(current->sound_filename[i]);
		}

		free(client_list[client]);
		if (client < num_clients-1)
			memmove(client_list+client, client_list+client+1, sizeof(void *) * (num_clients-1-client));
		return 0;
	}
	else return -1;
}

/*---------------------------------------------------------------------------*
 *	EOF clients.c
 *---------------------------------------------------------------------------*/
